####################################################################################################
# Examples using US data to illustrate how to use the various 'bnf' methods 
####################################################################################################

# Clear the workspace -- comment out if you have loaded other packages already
rm(list = ls(all = TRUE))
gc()

# Source required functions
source("bnf_methods.R")

# Read in US centric data to use for the demonstration
usdata <- read.csv(file = 'us_data.csv', header = TRUE, stringsAsFactors = FALSE)

# Make the series GDP a 'ts' object
# type 'help(ts)' in the R console for more information
gdp <- ts(data = usdata$GDPC1, end = c(2016, 4), frequency = 4)

#y <- transform_series(y = usdata$GDPC1, take_log = TRUE, pcode = "p1") # same as: log(raw_y) * 100.0
y <- transform_series(y = gdp, take_log = TRUE, pcode = "p1") # same as: log(raw_y) * 100.0

# Example 1: Automatically determined delta and rolling mean with window = 40 demeaning method
cat("\nExample 1: Automatically determined delta and rolling mean with window = 40 demeaning method\n\n")
window <- 40
bnf1 <- bnf(y = y, demean = "dm", wind = window)
plot(bnf1, main = sprintf("Cycle using \'Rolling-mean\' demeaning (window = %d)", window))

# Example 2: Automatically determined delta and full sample mean demeaning method
cat("Example 2: Automatically determined delta and full sample mean demeaning method\n\n")
bnf2 <- bnf(y)   # use default 'p', 'd0' and 'dt' values and "sm" demean
write.csv(bnf2$cycle, "us_cycle_R_bnf2.csv")
plot(bnf2, main = "US Output Gap", col = "red")
cat("\nPrinting out cycle data...\n")
print(bnf2) # comment this command to stop the cycle data being printed to the console
cat('\n')

# Example 3: Automatically determined delta and piecewise mean with 'breaks' = c(50, 75, 100, 125, 150) demeaning method
# NB: This example is only to show how one uses the piecewise_demean option.
cat("Example 3: Automatically determined delta and piecewise mean...\n\t\t...with 'breaks' = c(50, 75, 100, 125, 150, 175) demeaning method (demo only)\n\n")
bnf3 <- bnf(y = y, demean = "pm", breaks = c(50, 75, 100, 125, 150)) # if 'breaks' empty does 'sm'
plot(bnf3, main = "Cycle using \'Piecewise-mean\' demeaning", plot_ci = FALSE)

# Example 4: Manually estimate the cycle using the individual methods separately
cat("Example 4: Manually estimate the cycle using the individual methods separately\n\n")

# i) Pre-treat y: difference and demean GDPC1
dy <- transform_series(y = usdata$GDPC1, take_log = TRUE, dcode = "d1", pcode = "p1") # same as: diff(log(raw_y)) * 100.0
smdy <- scale(dy, center = TRUE, scale = FALSE)         # same as: y - mean(y)

u <- transform_series(y = na.omit(usdata$UNRATE), take_log = FALSE, dcode = "d1", pcode = "np")
smu <- scale(u, center = TRUE, scale = FALSE)           # same as: u - mean(u)

# Manually set number of lags to use in the BN filter step
# ii) Set 'p' and Estimate 'delta'
p <- 12 # bnf() uses 12 as the default value for 'p'
cat(sprintf("EG4: p manually set to: %d\n\n", p))

delta_y <- max_amp_to_noise(y = smdy, p = p)
delta_u <- max_amp_to_noise(y = smu, p = p)
cat(sprintf("EG4: GDP Delta manually set to: %2.2f\n\n", delta_y))
cat(sprintf("EG4: UR Delta manually set to: %2.2f\n\n", delta_u))

# iii) Estimate the cycle using the pre-treated version 'y' and the estimated 'delta'
# GDP
bnf4 <- BN_Filter(y = smdy, p = p, delta = delta_y)
gdp_cycle <- ts(data = bnf4$BN_cycle, end = c(2016, 4), frequency = 4)
write.csv(bnf4$BN_cycle, "us_cycle_R_bnf4.csv")          # to compare with bnf2

# Unemployment rate
bnfu <- BN_Filter(y = smu, p = p, delta = delta_u)
un_cycle <- ts(data = -bnfu$BN_cycle, end = c(2016, 4), frequency = 4)

# Combine ts data into a group for plotting
yu_cycle <- ts.intersect(gdp_cycle, un_cycle)
plot(yu_cycle, type = 'l', col = 4, lwd = 2, 
    main = "Output-Unemployment Cycle Comparison")

# EOF
